# coding=UTF-8
from operator import methodcaller
from skyeye.dto import *
from skyeye.result_wirter import ResultWirter
from .method_block_parser import InvokeLineParser
from skyeye.utils import *


class SmaliParser:

    #扫描单个类文件
    @classmethod
    def scanSingleSmali(cls,smaliFilePath='',scan_strategy_list=[]): 
        if(smaliFilePath.endswith(".smali") == False):
            return
        smaliFile = open(smaliFilePath)
        lines = smaliFile.readlines() 
  
        className = ""
        methodBlock = False
        matcherCallerInfo:CallerInfo = None
        for line in lines:
            # 解析调用类
            if(line.startswith('.class')):
                className = line[line.index('L')+1:line.index(";")].replace("/",".")
                continue
            # 解析调用方法
            if(line.startswith('.method')):
                 matcherCallerInfo = SmaliParser.parseCallMehtodInfo(line,className)
                 methodBlock = True
                 continue
             
            if(methodBlock):
                if(line.strip().startswith('.line')):
                    matcherCallerInfo.invoke_num = SmaliParser.parseInvokeNum(line)
                    continue  
                # 扫描调用方法里面 执行的方法  1.执行的方法 2.成员变量赋值的方法
                copyCallerInfo = InvokeLineParser.matchMethodBlock(matcherCallerInfo,line,scan_strategy_list)
                if(copyCallerInfo):
                    SmaliParser.toStringScanInfo(copyCallerInfo)
                    # 情况扫描信息
                    matcherCallerInfo.target_class = None
                    matcherCallerInfo.target_method = None
                    matcherCallerInfo.target_ref_filed = None
                    matcherCallerInfo.invoke_num = None
                    ResultWirter.shared().addResultDto(copyCallerInfo)
            # 解析调用方法结束
            if(line.startswith('.end method')):
                methodBlock = False
                matcherCallerInfo = None
        smaliFile.close()
        # print("扫描类 "+className)
        

    #调用者函数
    @classmethod
    def parseCallMehtodInfo(cls,startMethodLine='',className='')->CallerInfo:
        callerInfo = CallerInfo()
        calllMehotdInfoList = startMethodLine.split(' ')
        methodDsec = ''
        for index,item in enumerate(calllMehotdInfoList):
            if(index != 0):
                methodDsec += " "+item.strip().replace("/",".")
        callerInfo.caller_class = className
        callerInfo.caller_method = methodDsec.strip()
        return callerInfo


    #调用者执行的行数
    @classmethod
    def parseInvokeNum(cls,line:str)->str:
        lineStrip = line.strip()
        lineNumStr = lineStrip[5:]
        return lineNumStr


    @classmethod
    def toStringScanInfo(cls,copyCallerInfo:CallerInfo)->str:
        resultLine = ''
        if not isEmpty(copyCallerInfo.target_method):
            resultLine ='行数={e} {a}:{b} 调用: {c}:{d}'.format(a=copyCallerInfo.caller_class,
                                b=copyCallerInfo.caller_method,
                                c=copyCallerInfo.target_class,
                                d=copyCallerInfo.target_method,
                                e=copyCallerInfo.invoke_num)
        else:
            resultLine ='行数={e} {a}:{b} 调用: {c}.{d}'.format(a=copyCallerInfo.caller_class,
                    b=copyCallerInfo.caller_method,
                    c=copyCallerInfo.target_class,
                    d=copyCallerInfo.target_ref_filed,
                    e=copyCallerInfo.invoke_num) 
        print("😄😄😄扫描到了"+resultLine)
        
        
  
        




# 扫描配置
class ScanVO:
    class_name = ""
    method_name = ""
    filed_name = ""

    def __init__(self, class_name=None, method_name=None, filed_name=None):
        self.class_name = class_name
        self.method_name = method_name
        self.filed_name = filed_name
  

    @classmethod
    def from_json(cls, json_map=None):
        if json_map is not None:
            return ScanVO(class_name=json_map.get("className"),
                         method_name=json_map.get("methodName"),
                         filed_name=json_map.get("filedName"))
        return ScanVO()




# coding=UTF-8
import requests
from distutils.version import StrictVersion
from skyeye.config import VERSION

def checkVersions():
    url = "https://pypi.python.org/pypi/skyeye/json"
    versions = []
    try:
        data = requests.get(url).json()
        if(data == None):
            return
        releases = data["releases"]
        if(releases == None):
            return
        versions = sorted(list(releases.keys()), key=StrictVersion, reverse=True)
    except:
        pass
    if(versions and len(versions)>0):
        hasNewVersion = StrictVersion(versions[0]) > StrictVersion(VERSION)
        if hasNewVersion:
           print('🔥🔥🔥 版本升级提示，最新版本：'+versions[0]+" \n🔥🔥🔥请用该命令升级： pip3 install --upgrade skyeye")



# coding=UTF-8
import yaml
from skyeye.dto import *
import os

# 配置中心
class ConfigCenter:
    
    scan_config = []

    @classmethod
    def initialize(cls,configFilePath):
        ConfigCenter.shared().parseConfig(configFilePath)

    @classmethod
    def shared(cls):
        if not hasattr(ConfigCenter, "_instance"):
            ConfigCenter._instance = ConfigCenter()
        return ConfigCenter._instance

    def __init__(self):
        pass

    def parseConfig(self,configFilePath):
        realPath = os.path.join(os.getcwd(),configFilePath)
        f=open(realPath, 'r', encoding='utf-8')
        config_data = yaml.load(f.read(), Loader=yaml.FullLoader)
        for strategy in config_data:
            scanVo = ScanVO.from_json(strategy)
            self.scan_config.append(scanVo)


# coding=UTF-8
import copy
from skyeye.dto import *

import re

class InvokeLineParser:

    @classmethod
    def upperFirstChar(cls,str)->str:
        return str[0:1].upper()+ str[1:]
        
    @classmethod
    def matchMethodBlock(cls,matcherCallerInfo:CallerInfo,line:str,scan_strategy_list=[ScanVO])->CallerInfo:
        if None == matcherCallerInfo:
            return None
        lineStrip = line.strip()
        if(lineStrip.startswith('invoke')):
            return InvokeLineParser.matchInvoekLine(matcherCallerInfo,lineStrip,scan_strategy_list)
        if(lineStrip.startswith('iget-object') or lineStrip.startswith('sget-object')):
            return InvokeLineParser.matchFiled(matcherCallerInfo,lineStrip,scan_strategy_list)
        else:
            return  None

    #解析 静态变量引用(sget-object) 变量引用(iget-object)
    @classmethod
    def matchFiled(cls,matcherCallerInfo:CallerInfo,lineStrip,scan_strategy_list=[ScanVO])->CallerInfo:
        # sget-object v1, Lcn/cheney/picker/app/Test;->staticString:Ljava/lang/String;
        classMatch = re.compile(r",{1}\s\S+;->{1}").search(lineStrip)
        if not classMatch:
            return None
        tempClassStr = classMatch.group()
        getClass = tempClassStr[3:len(tempClassStr)-3].replace('/','.')

        filedMatch = re.compile(r"->{1}[\s\S]+:{1}").search(lineStrip)
        if not filedMatch:
            return None
        tempFiledStr = filedMatch.group()
        getFiled = tempFiledStr[2:len(tempFiledStr)-1].replace('/','.')
        for strategy in scan_strategy_list:
            targetClass = strategy.class_name
            targetFiled = strategy.filed_name
            if(targetClass in getClass and matcherCallerInfo.caller_class not in getClass):
                if(targetFiled == None or len(targetFiled) == 0):
                    matcherCallerInfo.target_class = getClass
                    matcherCallerInfo.target_ref_filed = getFiled
                    return copy.copy(matcherCallerInfo)
                else:
                    if getFiled in targetFiled and getClass in targetClass:
                         matcherCallerInfo.target_class = getClass
                         matcherCallerInfo.target_ref_filed = getFiled
                         return copy.copy(matcherCallerInfo)   
            else:
                return None
                
    
    #解析invoke里面的调用
    @classmethod
    def matchInvoekLine(cls,matcherCallerInfo:CallerInfo,lineStrip,scan_strategy_list=[])->CallerInfo:
        # invoke-static {}, Ljava/lang/Runtime;->getRuntime()Ljava/lang/Runtime;
        # 解析完成之后 invokeClass = java.lang.Runtime
        classMatch = re.compile(r"},{1}\s\S+;->{1}").search(lineStrip)
        if not classMatch:
            return None
        tempClassStr = classMatch.group()
        invokeClass = tempClassStr[4:len(tempClassStr)-3].replace('/','.')
        #解析完成之后 invokeMethod = getRuntime()
        methodMatch = re.compile(r"->{1}[\s\S]+").search(lineStrip)
        if not methodMatch:
            return None
        invokeMethod = methodMatch.group()[2:].replace('/','.')
        # print ("matchInvokeLine invokeClass="+invokeClass+" invokeMethod="+invokeMethod)
        for strategy in scan_strategy_list:
            targetClass = strategy.class_name
            targetMethod = strategy.method_name
            targetFiled = strategy.filed_name
            
            # 执行方法的类包含扫描的类&& 不包括自生的调用
            if(targetClass in invokeClass and matcherCallerInfo.caller_class not in invokeClass):
                # 匹配所有调用到目标类的invoke line
                if(targetMethod == None or len(targetMethod) == 0):
                    if (targetFiled == None or len(targetFiled) == 0):
                        matcherCallerInfo.target_class = targetClass
                        matcherCallerInfo.target_method = invokeMethod
                        return copy.copy(matcherCallerInfo)
                    else:
                        # 初始字段的set和get方法
                        unperFirstCharFiled = InvokeLineParser.upperFirstChar(targetFiled)
                        targetFiledToSetFunc = 'set'+unperFirstCharFiled
                        targetFiledToGetFunc = 'get'+unperFirstCharFiled
                        if(targetFiledToSetFunc in invokeMethod or targetFiledToGetFunc in invokeMethod):
                            matcherCallerInfo.target_class = targetClass
                            matcherCallerInfo.target_ref_filed = targetFiled
                            return copy.copy(matcherCallerInfo)
                # 匹配调用到目标类和目标方法的invoke line
                elif(targetMethod in invokeMethod):
                    matcherCallerInfo.target_class = targetClass
                    matcherCallerInfo.target_method = targetMethod
                    return copy.copy(matcherCallerInfo)
        return None



# coding=UTF-8
import os
from skyeye.dto import *
from skyeye.utils import *
import json

class ResultWirter:
    
    callerInfoList = [CallerInfo]
    
    callerResultMap = {}
    
    @classmethod
    def initialize(cls):
        ResultWirter.shared()

    @classmethod
    def shared(cls):
        if not hasattr(ResultWirter, "_instance"):
            ResultWirter._instance = ResultWirter()
        return ResultWirter._instance

    def __init__(self):
        if os.path.exists('out/result.txt'):
           os.remove('out/result.txt')

    
    def addResultDto(self,callerInfo):
        self.callerInfoList.append(callerInfo)
    
    def wirte(self,resultDir):
        self.groupCaller()
        jsonStr = json.dumps(self.callerResultMap,indent=4, ensure_ascii=False)
        realOutputPath = os.path.join(os.getcwd(),"result.json")
        if(resultDir):
            realOutputPath = os.path.join(os.getcwd(),resultDir,"result.json")
        with open(realOutputPath,"w+") as file:
             file.writelines(jsonStr)
        return realOutputPath   

    def groupCaller(self):
        for callerInfo in self.callerInfoList:
            key = ''
            if(not isEmpty(callerInfo.target_method)):
                key ='{a}:{b}'.format(a=callerInfo.target_class, b=callerInfo.target_method)
            elif(not isEmpty(callerInfo.target_ref_filed)):
                key ='{a}.{b}'.format(a=callerInfo.target_class, b=callerInfo.target_ref_filed)
            if(isEmpty(key)):
                continue
            callerList = self.callerResultMap.get(key)
            if callerList != None :
                callerList.append(CallerInfo.toResultJsonMap(callerInfo))
                continue
            callerList = []
            callerList.append(CallerInfo.toResultJsonMap(callerInfo))
            self.callerResultMap[key] = callerList



class CallerInfo:
    #调用者的类名
    caller_class = ''
    #调用者的方法
    caller_method = ''
    #目标扫描类名
    target_class = ''
    #目标扫描方法
    target_method = ''
    #目标扫描的成员变量
    target_ref_filed = ''
    #执行的行数
    invoke_num = ''

    def __init__(self, caller_class="", caller_method="", target_class="",target_method="", invoke_num="",target_ref_filed = ""):
        self.caller_class = caller_class
        self.caller_method = caller_method
        self.target_class = target_class
        self.target_method = target_method
        self.invoke_num = invoke_num
        self.target_ref_filed = target_ref_filed
    
    def toResultJsonMap(self):
        jsoMap = {}
        jsoMap['caller_class'] = self.caller_class
        jsoMap['caller_method'] = self.caller_method
        jsoMap['invoke_num'] = self.invoke_num
        return jsoMap
    


# coding=UTF-8
import os
import sys
import zipfile
import time
import shutil
from skyeye.config import ConfigCenter
from skyeye.result_wirter import ResultWirter
from skyeye.parse import SmaliParser

def startScan(apkPath,configYamlPath,resultDir):
    # 初始化配置
    ConfigCenter.initialize(configYamlPath)
    # 获取扫描规则 
    if(len(ConfigCenter.scan_config)==0):
        print("请配置扫描策略")
        return
    ResultWirter.initialize()
    # 清除之前可能遗留的临时文件
    delAllTempFiles()
    dexList = getDexFileList(apkPath)
    toSmail(dexList)
    if  not scanSmalis(ConfigCenter.scan_config) :
        delAllTempFiles()
        return
    resultJsonFilePath = ResultWirter.shared().wirte(resultDir)
    print("✅ 扫描结果: "+resultJsonFilePath)
    delAllTempFiles()

    
def scanSmalis(scan_strategy_list):
    print("🚀 开始扫描,过程大概持续2分钟...")
    startTime = int(time.time())
    outputSmaliDirPath = os.path.join(os.getcwd(),'out/smali')
    if not os.path.isdir(outputSmaliDirPath):
        print("目标smali文件夹="+outputSmaliDirPath+"为空，结束扫描❌")
        return False
    walk = os.walk(outputSmaliDirPath)
    for path,dir_list,file_list in walk:
        for file_name in file_list:
            SmaliParser.scanSingleSmali(os.path.join(path,file_name),scan_strategy_list)
    scanTime = int(time.time()) - startTime
    print("✅ 扫描耗时"+str(scanTime)+"s")
    return True
           
#  将dex转换成smail           
def toSmail(dexList):
    startTime = int(time.time())
    print("🚀 换成smali,过程大概持续2分钟...")
    current_file_dir = os.path.dirname(__file__)
    outputSmaliDirPath = os.path.join(os.getcwd(),'out/smali')
    if os.path.isdir(outputSmaliDirPath) == True:
       shutil.rmtree(outputSmaliDirPath)
    for dexPath in dexList:
        baksmilJarPath=  os.path.join(current_file_dir,'config/baksmali-2.5.2.jar')
        # baksmilJarPath= os.path.abspath('/config/baksmali-2.5.2.jar')
        command = "java -jar "+ baksmilJarPath+" d "+dexPath+" -o "+outputSmaliDirPath
        toSmailResult = os.popen(command).readlines()
        # print("换成smali----toSmailResult="+str(toSmailResult))
    toSmailTime = int(time.time()) - startTime
    # print("✅ 换成smali成功 耗时"+str(toSmailTime)+"s")

# 提取APK中的dex文件
def getDexFileList(apkPath):
    realApkPath = os.path.join(os.getcwd(),apkPath)
    outputDexDirPath = os.path.join(os.getcwd(),"out/dex")
    apkFile = zipfile.ZipFile(realApkPath,"r")
    dexFileList = []
    
    if os.path.exists(outputDexDirPath):
       shutil.rmtree(outputDexDirPath)
    if os.path.isdir(outputDexDirPath) == False:
       os.makedirs(outputDexDirPath)

    for tempFile in apkFile.namelist():
        if tempFile.endswith(".dex"):
            dexFilePath = os.path.join(outputDexDirPath,tempFile)
            dexfile = open(dexFilePath,'wb+')
            dexfile.write(apkFile.read(tempFile))
            dexFileList.append(dexFilePath)
            dexfile.close()
        
    return dexFileList

def delAllTempFiles():
    startTime = int(time.time())
    outputDirPath = os.path.join(os.getcwd(),"out")
    if os.path.exists(outputDirPath):
       print("🚀 删除临时文件夹"+outputDirPath+" ...")
       shutil.rmtree(outputDirPath)
       delDirTimes = int(time.time()) - startTime
       print("✅ 删除临时文件夹"+str(delDirTimes)+"s")
  



#-*- encoding: UTF-8 -*-
from setuptools import setup,find_packages
import io
from skyeye.config import VERSION


with io.open("README.md", 'r', encoding='utf-8') as f:
    long_description = f.read()
    
setup(name='skyeye',
      version= VERSION,
      description="APK扫描工具",
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='apk scanner',
      author='wangshuwen',
      author_email='wnwn7375@outlook.com',
      url='https://github.com/wangshuwen1107/skyeye',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
        'requests',
        'PyYaml'
      ],
      entry_points={
        'console_scripts':[
            'skyeye = skyeye.cli:run'
        ]
      },
)





# coding=UTF-8
import argparse
from skyeye.apk_scanner import startScan
from skyeye.config import VERSION
from skyeye.version_check import checkVersions

def version():
    return "version:"+VERSION


def scan(args):
    inputApkPath = args.input
    configYamlPath = args.config
    outputResultDirPath = args.output
    startScan(inputApkPath,configYamlPath,outputResultDirPath)


def run():
    parser = argparse.ArgumentParser(prog="skyeye")

    parser.add_argument("-v","--version", help=u"查看版本号",action="store_true")
    
    subparsers = parser.add_subparsers(title='扫描命令',metavar="scan",help="扫描APK中的静态引用")
    scan_parser = subparsers.add_parser('scan')
    scan_parser.add_argument("-i","--input", type=str, help="输入扫描的APK路径",metavar="")
    scan_parser.add_argument("-c","--config", type=str, help="扫描配置yaml文件路径",metavar="")
    scan_parser.add_argument("-o","--output", type=str,nargs='?', help="输出结果的文件夹路径",metavar="")
    scan_parser.set_defaults(func=scan)
    
    args = parser.parse_args()
    # print("args="+str(args))

    if hasattr(args, "func"):
        args.func(args)
        return
    if args.version:
        checkVersions()
        print(version())
        return
    parser.print_help()
       
       
       
if __name__ == '__main__':
    run()